# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""A time tracker for the command line. Utilizing the power of nark."""

from gettext import gettext as _

import copy
import re
import sys
from datetime import datetime, time, timedelta

from nark.helpers.parsing import parse_factoid

from dob_bright.termio import (
    attr,
    barf_and_exit,
    bg,
    click_echo,
    echo_block_header,
    fg
)

from .fact_dressed import FactDressed
from .fix_times import (
    DEFAULT_SQUASH_SEP,
    mend_facts_times,
    must_complete_times,
    reduce_time_hint,
    then_extend_fact
)
from .parse_mistakes import prepare_log_msg

__all__ = (
    'parse_input',
)


def parse_input(controller, file_in=None, progress=None):
    """
    Import Facts from STDIN or a file.
    """

    # MAYBE/2018-05-16 00:11: (lb): Parse whole file before prompting.
    #                               Allow --yes to work here, too?
    #                               Not sure...
    #   For now, we'll just read one fact at a time.

    def _parse_input():
        raw_facts = parse_facts_input(file_in)
        new_facts = must_hydrate_facts(raw_facts)
        conflicts = must_complete_times(controller, new_facts, progress=progress)
        controller.affirm(not conflicts)  # (lb): 2019-01-19: This happen?
        repair_shoulder_fact_times(new_facts, raw_facts)
        must_not_conflict_existing(new_facts)
        return new_facts

    # ***

    def parse_facts_input(file_in):
        # A few reminders:
        # - On `dob import`, file_in is None -- but on `dob import -`,
        #   Click sets file_in to <stdin> _io.TextIOWrapper.
        # - On `echo '...' | dob import` → isatty() is False, but
        #   on `dob import | cat`        → isatty() is True.
        # - Dob will print help on `dob import`, but if the user runs
        #   `dob import -` then dob waits for input from the user, and
        #   processes it line by line (whenever the user his ENTER).
        #   This is expected behavior -- that's what the special `-`
        #   filename convention does in *nix -- but I thought I'm spell
        #   it out, in case it feels odd to new developers.
        input_f = file_or_stdin(file_in)
        raw_facts = parse_facts_from_stream(input_f)
        return raw_facts

    def file_or_stdin(file_in):
        if file_in is not None:
            return file_in
        return sys.stdin

    def parse_facts_from_stream(input_f):
        # Track current empty line run count, to know when to check if line
        # starts a new Fact. Keep at -1 until we've seen the first fact.
        bl_count = -1

        line_num = 0

        # Coalesce each Fact, line by line.
        current_fact_dict = None
        accumulated_fact = []
        unprocessed_facts = []

        progress and progress.click_echo_current_task(_('Parsing factoids...'))
        for line in input_f:
            line_num += 1
            (
                bl_count, processed,
            ) = gobble_blank_or_continued(line, accumulated_fact, bl_count)
            if processed:
                continue
            fact_dict = gobble_if_not_new_fact(line, accumulated_fact, bl_count)
            if fact_dict is None:
                bl_count = 0
                continue
            fact_dict['parsed_source.line_num'] = line_num
            fact_dict['parsed_source.line_raw'] = line
            if not accumulated_fact:
                # First Fact.
                assert current_fact_dict is None
                assert not accumulated_fact
                current_fact_dict = fact_dict
                accumulated_fact.append(line)
                bl_count = 0
                continue
            else:
                assert current_fact_dict is not None
                assert accumulated_fact

            # Woot, woot! We parsed a complete Fact.
            if current_fact_dict:
                unprocessed_facts.append((current_fact_dict, accumulated_fact,))
            current_fact_dict = fact_dict
            accumulated_fact = [line, ]
            bl_count = 0

        # end: for

        if accumulated_fact:
            unprocessed_facts.append((current_fact_dict, accumulated_fact,))
        else:
            msg = _('What is this, an empty file?')
            barf_and_exit(msg)

        return unprocessed_facts

    def gobble_blank_or_continued(line, accumulated_fact, blank_line_count):
        processed = False
        if blank_line_count < 0:
            # Start of file.
            assert blank_line_count == -1
            if not line.strip():
                controller.client_logger.debug(_('Skip premature blank line'))
                # Skip blank lines pre-Facts.
                processed = True
            # else, we found the first non-blank line. We now
            #   expect to find the date:time & meta, or death.
        elif not line.strip():  # remove trailing newline
            controller.client_logger.debug(_('- Blank line in desc.'))
            accumulated_fact.append(line)
            blank_line_count += 1
            processed = True
        elif blank_line_count == 0:
            controller.client_logger.debug(_('- Part of desc.:\n') + line.strip())
            accumulated_fact.append(line)
            processed = True
        elif re.match(r'^\s', line):
            controller.client_logger.debug(_('- Leading whitesp.:\n') + line.strip())
            accumulated_fact.append(line)
            processed = True
        # else, any content that follows blank line(s) is either:
        # (1) more content; (2) a new Fact; or (3) Fact separator.
        return (blank_line_count, processed)

    def gobble_if_not_new_fact(line, accumulated_fact, blank_line_count):
        fact_dict = dissect_meta_line(line)
        if fact_dict is None:
            # If not parsed, and first line we've seen, die.
            missing_fact_must_come_early(line, blank_line_count)
            # else, more content, or a Fact separator.
            controller.client_logger.debug(_('- More desc.: ') + line.strip())
            accumulated_fact.append(line)
        return fact_dict

    def missing_fact_must_come_early(line, blank_line_count):
        if blank_line_count == -1:
            msg = (
                '{}: {}{}{}{}{}\n'
                .format(
                    _('The first nonempty line is not a recognized Fact.'),
                    bg('white'), fg('black'), attr('underlined'),
                    line.strip(),
                    attr('reset'),
                )
            )
            barf_and_exit(msg)

    # FIXME/2019-01-21: Document all the different usage, both in test, and README.
    # E.g., "then: I did this" should be same as "at +0: I did this"
    #   but if you specify start you don't need colon
    #       "then 2019-01-21 23:47 I did this" should also work...
    #   FIXME: Parse "then" expecting 1+ datetimes, or "then:" expecting none.
    #   FIXME: Parse "still" expecting 1+ datetimes, or "still:" expecting none.
    # FIXME: Three tests:
    #           ``still <time-spec> <desc>``
    #       vs. ``still: <desc>``
    #       vs. ``still blah``
    RE_TIME_HINT = re.compile(
        # SYNC_ME: RE_TIME_HINT, TIME_HINT_MAP, and @generate_add_fact_command's.
        r'^('
            # Skipping: Doesn't make sense: '(?P<verify_none>on|now)'
            # MAYBE/2019-01-22: Is "between" okay here?
            #   We don't have a Click alias for it, so
            #   there's a `dob from` command, but not `dob between`,
            #   so maybe we want to remove "between" from here.
            '(?P<verify_both>from|between)'
            '|(?P<verify_start>at)'  # noqa: E131
            '|(?P<verify_end>to|until)'
            '|(?P<verify_then_none>then:)'
            '|(?P<verify_then_some>then)'
            '|(?P<verify_still_none>still:)'
            '|(?P<verify_still_some>still)'
            # NOTE: Require colon postfix for options w/o time component.
            # NOTE: 'now' would be confusing and conflict with other usage,
            #       (at least I think it would?). E.g., do not do this:
            #         '|(?P<verify_after>after:|since:|next:|now:)'
            '|(?P<verify_after>after:|since:|next:)'
        ' )',  # NOTE The SPACE CHARACTER following THE DIRECTIVE!
        re.IGNORECASE,
    )

    def dissect_meta_line(line):
        sussed_hint, time_hint, line = suss_the_time_format(line)
        fact_dict, err = run_parser(line, time_hint)
        if not fact_dict['start'] and not fact_dict['end']:
            controller.client_logger.debug(_('unmeta: {}'.format(err)))
            fact_dict = None
        else:
            # The user could specify, e.g., just a single datetime (followed
            # by a description), and we can fill in the other datetime and
            # ask the user for more details about the fact.
            #   Not always True:  assert not err
            controller.client_logger.debug(_(
                # Including the description is too verbose:
                #   'new fact_dict: {}'.format(fact_dict)
                # so let's make a complicated dictionary comprehension instead.
                'new fact_dict: {}'.format(
                    {
                        key: val if key != 'description'
                        else val[:10] + ((len(val) > 10) and '...' or '')
                        for key, val in fact_dict.items()
                    }
                )
            ))
        return fact_dict

    def suss_the_time_format(line):
        sussed_hint = ''
        match = RE_TIME_HINT.match(line)
        if match is not None:
            for hint, matched in match.groupdict().items():
                if matched is not None:
                    assert not sussed_hint
                    sussed_hint = hint
                    # Remove the time hint prefix.
                    line = RE_TIME_HINT.sub('', line).lstrip()
        time_hint = sussed_hint or 'verify_start'
        controller.client_logger.debug(_(
            'time_hint: {}{}'.format(
                time_hint, ' [sussed]' if sussed_hint else ' [default]',
            )
        ))
        return sussed_hint, time_hint, line

    def run_parser(line, time_hint):
        # Parse the line as if it were a new fact. And don't strip the line.
        # Parser expects metadata to be separated from description (so leave
        # the newline). And we can be strict and require that the date data
        # starts the line.

        # Per dob-insert commands, parser expects iterable.
        factoid = (line,)

        use_hint = reduce_time_hint(time_hint)

        fact_dict, err = parse_factoid(
            factoid=factoid,
            time_hint=use_hint,
            # MEH: (lb): Should we leave hash_stamps='#@' ?
            #   I sorta like using the proper tag symbol
            #   when not worried about shell interpolation.
            hash_stamps='#',
            # We'll handle errors ourselves, later, in bulk, either
            # `--ask`'ing the user for more Fact details, or barfing
            # all the errors and exiting.
            lenient=True,
        )

        fact_dict_set_time_hint(fact_dict, time_hint)

        return fact_dict, err

    # FIXME/DRY: See create.py/transcode.py (places that use "+0").
    def fact_dict_set_time_hint(fact_dict, time_hint):
        fact_dict['time_hint'] = time_hint
        if time_hint in ('verify_after', 'verify_then_none', 'verify_still_none'):
            assert not fact_dict['start'] and not fact_dict['end']
            # (lb): How's this for a hack!?
            fact_dict['start'] = "+0"

    # ***

    def must_hydrate_facts(raw_facts):
        new_facts, hydrate_errs = hydrate_facts(raw_facts)
        must_hydrated_all_facts(hydrate_errs)
        return new_facts

    def hydrate_facts(raw_facts):
        new_facts = []
        hydrate_errs = []
        temp_id = -1
        progress and progress.click_echo_current_task(_('Hydrating Facts...'))
        for fact_dict, accumulated_fact in raw_facts:
            add_hydration_warnings(fact_dict, hydrate_errs)
            hydrate_description(fact_dict, accumulated_fact)
            new_fact, err_msg = create_fact_from_parsed_dict(fact_dict)
            if new_fact:
                assert not err_msg
                time_hint = fact_dict['time_hint']
                temp_id = add_new_fact(new_fact, time_hint, temp_id, new_facts)
            else:
                assert not new_fact
                hydrate_errs.append(err_msg)
        return new_facts, hydrate_errs

    def add_hydration_warnings(fact_dict, hydrate_errs):
        if not fact_dict['warnings']:
            return
        fact_warnings = '\n  '.join(fact_dict['warnings'])
        display_dict = copy.copy(fact_dict)
        del display_dict['warnings']
        err_msg = prepare_log_msg(display_dict, fact_warnings)
        hydrate_errs.append(err_msg)

    # Horizontal rule separator matches same character repeated at least thrice.
    # FIXME/2018-05-18: (lb): Document: HR is any repeated one of -, =, #, |.
    FACT_SEP_HR = re.compile(r'^([-=#|])\1{2}\1*$')

    def hydrate_description(fact_dict, accumulated_fact):
        # The first entry is the meta line, so ignore it.
        desc_lines = accumulated_fact[1:]
        # However, the user is allowed to start the description
        # on the meta line.
        if fact_dict['description']:
            # Don't forget the newline -- it got culled (strip()ped) on parse.
            desc_lines.insert(0, fact_dict['description'] + "\n")

        cull_factless_fact_separator(desc_lines)

        # desc_lines are already newlined.
        desc = ''.join(desc_lines).strip()

        fact_dict['description'] = desc

    def cull_factless_fact_separator(desc_lines):
        # To make import file more readable, user can add
        # separator line between facts. Cull it if found.
        # MAYBE/2018-05-18: (lb): Make this operation optional?
        while len(desc_lines) > 0 and not desc_lines[-1].strip():
            desc_lines.pop()

        if (
            len(desc_lines) > 1
            and not desc_lines[-2].strip()
            and FACT_SEP_HR.match(desc_lines[-1])
        ):
            # Remove optional horizontal rule, which user
            # can use to visual separate lines in import file.
            desc_lines.pop()

        return desc_lines

    def create_fact_from_parsed_dict(fact_dict):
        new_fact = None
        err_msg = None
        try:
            new_fact = FactDressed.create_from_parsed_fact(
                fact_dict,
                lenient=True,
                line_num=fact_dict['parsed_source.line_num'],
                line_raw=fact_dict['parsed_source.line_raw'],
            )
        except ValueError as err:
            err_msg = prepare_log_msg(fact_dict, str(err))
        # (lb): Returning a tuple that smells like Golang: (fact, err, ).
        return new_fact, err_msg

    def add_new_fact(new_fact, time_hint, temp_id, new_facts):
        if new_fact is None:
            return temp_id
        add_facts = maybe_squash_extend_prev(new_fact, time_hint, new_facts)
        for add_fact in add_facts:
            temp_id = add_new_fact_maybe(add_fact, temp_id, new_facts)
        return temp_id

    def maybe_squash_extend_prev(new_fact, time_hint, new_facts):
        if time_hint not in [
            'verify_end',
            'verify_then_none',
            'verify_then_some',
            'verify_still_none',
            'verify_still_some',
        ]:
            return [new_fact]
        return squash_extend_if_prev(new_fact, time_hint, new_facts)

    def squash_extend_if_prev(new_fact, time_hint, new_facts):
        prev_fact = new_facts[-1] if new_facts else None
        new_new_facts = squend_check_noop_or_extend(prev_fact, new_fact, time_hint)
        if new_new_facts is not None:
            return new_new_facts
        return squend_squash_or_extend(prev_fact, new_fact, time_hint, new_facts)

    def squend_check_noop_or_extend(prev_fact, new_fact, time_hint):
        if prev_fact is None:
            # First fact in import. Will check store later for boundary fact.
            return [new_fact]
        elif prev_fact.end is not None:
            return maybe_set_start_maybe_extend_prev(
                prev_fact, new_fact, time_hint,
            )
        elif prev_fact.start is None:
            return [new_fact]
        elif new_fact.start and new_fact.end:
            # This fact is a `then ... to`, so really just a `from/between`.
            return [new_fact]
        # E.g., prev_fact is at-fact, with no end; and new_fact is to-fact,
        # with no start. Return None, and caller will squend_squash_or_extend.
        assert not new_fact.deleted
        return None

    def squend_squash_or_extend(prev_fact, new_fact, time_hint, new_facts):
        squash_facts_maybe(new_fact, prev_fact)
        if not new_fact.deleted:
            return [new_fact]

        prev_or_new_fact = maybe_extend_fact(controller, prev_fact, time_hint)
        # Return None is prev fact was extended.
        if prev_or_new_fact is not new_fact:
            assert prev_or_new_fact is new_facts[-1]
        return [prev_or_new_fact] if prev_or_new_fact is new_fact else []

    def maybe_set_start_maybe_extend_prev(prev_fact, new_fact, time_hint):
        if time_hint in [
            'verify_then_none',
            'verify_then_some',
            'verify_still_none',
            'verify_still_some',
        ]:
            extended_fact = then_extend_fact(controller, prev_fact)
            extended_fact.end = new_fact.start
            if time_hint in ['verify_still_none', 'verify_still_some']:
                new_fact.activity = prev_fact.activity
                new_fact.tags = list(prev_fact.tags)
            return [extended_fact, new_fact]
        else:
            assert time_hint == 'verify_end'  # ``to``
            assert new_fact.start is None
            # Fix it in post, er, fix_delta_time_relative.
            new_fact.start = "+0"
            return [new_fact]

    def squash_facts_maybe(new_fact, prev_fact):
        assert new_fact.pk is None
        # 2018-06-30: (lb): Ug. Why does this squash logic seem so forced?
        new_fact.pk = -1  # To appease squash().
        prev_fact.squash(new_fact, DEFAULT_SQUASH_SEP)
        new_fact.pk = None

    def maybe_extend_fact(controller, new_fact, time_hint):
        if time_hint not in ['verify_then_none', 'verify_then_some']:
            return new_fact
        return then_extend_fact(controller, new_fact)

    def add_new_fact_maybe(new_fact, temp_id, new_facts):
        if not new_fact or new_fact.deleted:
            return temp_id
        new_fact.pk = temp_id
        new_facts.append(new_fact)
        temp_id -= 1
        return temp_id

    # ***

    def must_hydrated_all_facts(hydrate_errs):
        if not hydrate_errs:
            return
        for err_msg in hydrate_errs:
            click_echo()
            click_echo(err_msg)
            click_echo()
        msg = (_(
            'Please fix your import data and try again.'
            ' Scroll up for details.'
        ))
        barf_and_exit(msg)

    # ***

    def repair_shoulder_fact_times(new_facts, raw_facts):
        # If user did not specify first fact start time, and/or last fact's
        # end time, fix_times.must_complete_times will have assigned times.
        # Note that there are a number of places where two facts may be merged:
        # - Earlier during this import task, under add_new_fact(), squash and
        #   extend code is used.
        # - Here, but just on the first and final import facts, to sew them
        #   into the fabric of the facts in the store.
        # - In many of the fix_times functions:
        #     unite_and_stretch, insert_forcefully, and resolve_overlapping.
        # (lb): Maybe someday we can combine this code better. The way different
        #   features were grown, however, resulted in squash code in a few places.

        if not new_facts[0].start:
            # User did not specify first fact's start, and we did not find
            # an antecedent fact in the data store, so, what? Set earliest
            # time ever? Clip to midnight that day?
            new_facts[0].start = datetime.combine(new_facts[0].end, time.min)
            if new_facts[0].start == new_facts[0].end:
                new_facts[0].start = new_facts[0].end - timedelta(days=1)
        elif not raw_facts[0][0]['start']:
            new_facts[0].start = None

        controller.affirm(new_facts[-1].end)
        if not raw_facts[-1][0]['end']:
            new_facts[-1].end = None

    # ***

    def must_not_conflict_existing(new_facts):
        # (lb): Yuck. Sorry about this. Totally polluting what was a small
        # function with lots of progress-output overhead.
        task_descrip = _('Verifying times nonconflicting')
        if progress is not None:
            term_width, dot_count, fact_sep = progress.start_crude_progressor(
                task_descrip,
            )

        could_be_more = fix_range_conflicts_easy(new_facts)

        all_conflicts = []
        for idx, fact in enumerate(new_facts):
            if progress is not None:
                term_width, dot_count, fact_sep = progress.step_crude_progressor(
                    task_descrip, term_width, dot_count, fact_sep,
                )

            if idx == len(new_facts) - 1:
                time_hint = 'verify_last'
            elif idx == 0 and not fact.start:
                time_hint = 'verify_end'
            else:
                time_hint = 'verify_both'

            conflicts = mend_facts_times(
                controller,
                fact,
                time_hint=time_hint,
                skip_store=not could_be_more,
            )
            assert not fact.deleted  # Only on squash, which shouldn't happen.

            if conflicts:
                all_conflicts.append((fact, conflicts,))

        progress and progress.click_echo_current_task('')

        # Handling anything other than the ongoing fact seems beyond the
        # scope of the import function. dob is not a conflict resolution
        # application!
        barf_on_overlapping_facts_old(all_conflicts)

    def fix_range_conflicts_easy(new_facts):
        """
        Check time range of facts being imported against store.
        Returns True if time range overlaps with any existing facts;
        the ongoing fact only overlaps if it starts during the time
        range.
        """
        def _fix_range_conflicts_easy():
            first_time = new_facts[0].start or new_facts[0].end
            final_time = new_facts[-1].end or new_facts[-1].start
            # There's a compelling function, controller.facts.strictly_during,
            # but it ignores overlapping shoulder facts (and the ongoing fact).
            # Check instead manually against the fact before the last time,
            # and also the fact after the first time.
            ante_fact = controller.facts.antecedent(ref_time=final_time)
            seqt_fact = controller.facts.subsequent(ref_time=first_time)
            check_all = False
            check_all = easy_fix_final(check_all, seqt_fact, final_time)
            check_all = easy_fix_first(check_all, ante_fact, first_time)
            return check_all

        def easy_fix_final(check_all, seqt_fact, final_time):
            if not seqt_fact:
                return check_all
            if seqt_fact.start < final_time:
                # Fact found in store ends after first_time,
                #  and starts before final_time.
                return True
            if new_facts[-1].end:
                # The final import fact ends on or before
                #  the start of the after-fact.
                return check_all
            # If here, the final import fact is endless.
            # Obvious path is to end it using start of first fact after.
            if new_facts[-1].start == seqt_fact.start:
                # The ongoing final import fact starts at same time as final store fact.
                if not seqt_fact.end:
                    # To test this path: have ongoing fact in import with same start as
                    # ongoing fact in the store.
                    # (lb): We could end the existing fact, making it momentaneous,
                    # but that seems extreme; feels more natural to combine the 2
                    # facts. -- Alternatively, we could toss an error. But I could
                    # see just combining the 2 facts, seems like a pretty rare case
                    # anyway.
                    seqt_fact.squash(new_facts[-1], DEFAULT_SQUASH_SEP)
                    new_facts[-1] = seqt_fact
                    return check_all
                if not controller.config['time.allow_momentaneous']:
                    # User is not allowing momentaneous Facts, so disallow.
                    # Return True and we'll run thorough conflict check
                    # (and exit on error instead of continuing import).
                    return True
                # The final import fact is endless, but it conflicts with a
                # fact in the store. And the user is down with momentaneous
                # so crown it as such.
                # FIXME/2019-01-20 01:09: TESTME: Momentaneous feature.
                #  (I think it works in db and code, but no way for user
                #  to view in the UI. So they're essentially really well
                #  hidden. Which is a design feature, maybe?)
            new_facts[-1].end = seqt_fact.start
            return check_all

        def easy_fix_first(check_all, ante_fact, first_time):
            if not ante_fact:
                return check_all
            if ante_fact.end:
                if not new_facts[0].start:
                    # I.e., a to-Fact is 1st Fact in import,
                    #   and final Fact in store is closed.
                    new_facts[0].start = ante_fact.end
                if ante_fact.end > first_time:
                    # Fact from store starts before final_time,
                    #  and ends after first_time.
                    return True
                # else, first fact found before final_time
                #        ends before first_time.
                return check_all
            # If here, it's the ongoing fact.
            ongoing = controller.facts.endless()
            controller.affirm(ante_fact == ongoing[0])
            if new_facts[0].start:
                ante_fact.end = new_facts[0].start
                new_facts.insert(0, ante_fact)
            else:
                # Squash store's ongoing and import's first.
                ante_fact.squash(new_facts[0], DEFAULT_SQUASH_SEP)
                new_facts[0] = ante_fact

            return check_all

        return _fix_range_conflicts_easy()

    def barf_on_overlapping_facts_old(conflicts):
        if not conflicts:
            return

        for fact, resolved_edits in conflicts:
            for edited_fact, original in resolved_edits:
                echo_saved_fact_conflict(fact, edited_fact, original)
        msg = _(
            'One or more new Facts would conflict with existing Facts.'
            ' This Is Not Allowed'
        )
        barf_and_exit(msg)

    def echo_saved_fact_conflict(fact, edited_fact, original):
        echo_block_header(_('Saved Fact Datetime Conflict!'))
        click_echo()
        click_echo(_('  Fact being imported:'))
        click_echo(_('  ───────────────────'))
        click_echo(fact.friendly_diff(fact, truncate=True))
        click_echo()
        click_echo(_('  Impact on saved Fact:'))
        click_echo(_('  ────────────────────'))
        click_echo(original.friendly_diff(edited_fact))
        click_echo()

    # ***

    return _parse_input()




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Facts Carousel"""

import time
from datetime import timedelta

from gettext import gettext as _

from .exceptions import catch_action_exception
from .zone_content import ZoneContent

__all__ = (
    'UpdateHandler',
)


class UpdateHandler(object):
    """"""
    def __init__(self, carousel):
        self.carousel = carousel

        self.last_time_time_adjust = None
        self.press_cnt_time_adjust = 0

    # ***

    def standup(self):
        self.edits_manager = self.carousel.edits_manager
        self.zone_manager = self.carousel.zone_manager

    # ***

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def undo_command(self, event):
        """"""
        undone = self.edits_manager.undo_last_edit()
        if not undone:
            hot_notif = _("Already at oldest change")  # Like Vim says!
        else:
            self.zone_manager.rebuild_viewable()
            # FIXME: Do not exceed terminal width.
            #  hot_notif = _("Restored: {0}").format(edited_fact.short)
            hot_notif = _("Undid last change. Press Ctrl-y to redo")
        self.zone_manager.update_status(hot_notif)

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def redo_command(self, event):
        """"""
        redone = self.edits_manager.redo_last_undo()
        if not redone:
            hot_notif = _("Already at newest change")  # Same says Vim.
        else:
            self.zone_manager.rebuild_viewable()
            # FIXME: Do not exceed terminal width.
            #  hot_notif = _("Restored: {0}").format(edited_fact.short)
            hot_notif = _("Redid last change. Press Ctrl-z to undo")
        self.zone_manager.update_status(hot_notif)

    # ***

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_split(self, event):
        """"""
        self.carousel.controller.affirm(False)  # FIXME: Implement.

        self.zone_manager.rebuild_viewable()
        pass  # FIXME: Implement
        # FIXME/2019-01-15 13:51: For undo/redo, you'll need to leave
        # copy of new split Fact with copies of original Fact's
        # prev and next fact pointers, so that wires do not get crossed.

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_erase(self, event):
        """"""
        # FIXME: Implement with undo, or provide other way to recover.
        self.carousel.controller.affirm(False)  # FIXME: Implement.
        self.zone_manager.rebuild_viewable()
        pass  # FIXME: Implement

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_cut(self, event):
        """"""
        self.carousel.controller.affirm(False)  # FIXME: Implement.
        self.zone_manager.rebuild_viewable()
        pass  # FIXME: Implement

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_merge_prev(self, event):
        """"""
        self.carousel.controller.affirm(False)  # FIXME: Implement.
        self.zone_manager.rebuild_viewable()
        pass  # FIXME: Implement

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_merge_next(self, event):
        """"""
        self.carousel.controller.affirm(False)  # FIXME: Implement.
        self.zone_manager.rebuild_viewable()
        pass  # FIXME: Implement

    # ***

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_copy_activity(self, event):
        """"""
        self.edits_manager.fact_copy_activity()
        self.zone_manager.update_status(_("Copied activity"))

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_copy_tags(self, event):
        """"""
        self.edits_manager.fact_copy_tags()
        self.zone_manager.update_status(_("Copied tags"))

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_copy_description(self, event):
        """"""
        self.edits_manager.fact_copy_description()
        self.zone_manager.update_status(_("Copied description"))

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_copy_fact(self, event):
        """"""
        self.edits_manager.fact_copy_fact()
        self.zone_manager.update_status(_("Copied fact"))

    # ***

    @catch_action_exception
    @ZoneContent.Decorators.reset_showing_help
    def fact_paste(self, event):
        """"""
        def _fact_paste():
            paste_what = self.edits_manager.paste_copied_meta()
            hot_notif = paste_response(paste_what)
            self.zone_manager.update_status(hot_notif)

        def paste_response(paste_what):
            if not paste_what:
                return _("Nothing copied, nothing pasted")
            else:
                self.zone_manager.rebuild_viewable()
                return _("Pasted {0}").format(_(paste_what))

        _fact_paste()

    # ***

    # FIXME: (lb): These time +/- bindings are pretty cool,
    #        but they should and/or create conflicts,
    #        or affect the time of adjacent facts...
    #        so you may need to show times of adjacent facts,
    #        and/or you could auto-adjust adjacent facts' times,
    #        perhaps not letting user decrement below time of fact
    #        before it, or increment above time of fact after it.
    #        (at least need to preload the before and after facts for
    #        the current fact).
    #        for now, you can raw-edit factoid to change type
    #        and resolve conflicts. or you could edit facts in
    #        carousel and resolve conflicts yourself, before saving.
    #
    #        maybe if you begin to edit time, then you show
    #        adjacent fact's (or facts') time(s).
    #
    #        I think easiest option is to just adjust adjacent
    #        facts' times -- without showing them -- and to
    #        stop at boundary, i.e., when adjacent fact would
    #        be at 0 time (albeit, when if seconds don't match?
    #        you should normalize and remove seconds when you
    #        you adjust time!)
    #    1.  - normalize to 0 seconds on adjust
    #    2.  - change adjacent facts' time(s), too
    #    3.  - stop at adjacent fact boundary
    #    4.  - option to delete fact -- so user could either change to
    #          adjacent fact and edit its time to make more room for
    #          fact being re-timed; or user could switch to adjacent
    #          fact and delete it, poof!
    #    5.  - undo/redo stack of edit_facts
    #          e.g., undo delete fact command, or undo any edit:
    #             act@cat, tags, description!
    #             keep a copy of undo fact in carousel
    #    6.  - insert new fact -- or split fact! split fact in twain...
    #          YES! you could even have a split fact command, and then
    #          a clear fact command ... (s)plit fact... (C)lear fact...
    #          or maybe (S) and (C) ... this makes the redo/undo trickier...
    #          perhaps each redo/undo state is a copy of edit_facts?
    #          i think the trickier part is coordinating new_facts
    #          and edit_facts -- the lookup can return multiple facts,
    #          i'm guessing... and then the new split facts will just
    #          show the diff each against the same base fact... and on
    #          save... what? they all have the same split_from, I guess...
    #          because we don't really use split_from yet... OK, this is
    #          doable, super doable... and it could make my fact entering
    #          at end of day easier? or fact cleanup, i should mean, e.g.,
    #          if I had a fact that was 6 hours long, I could split it
    #          in two, and then adjust the time quickly with arrow keys!
    #    7.  - (h)elp option? TOGGLE HELP IN DESCRIPTION BOX!! You could
    #          probably easily redraw the header, too, to not show a fact...
    #          maybe use Header and Description area -- can I use a new
    #          Layout item or whatever from PPT and swap out current
    #          display? OR, leave the Fact header, because user might
    #          want to use one of the HELP KEYS while VIEWING THE HELP.
    #          So show help, and on any command, clear the help!
    #          And does this mean I should add more meta to the key bindings
    #          and build the help automatically? Or should I, say, just
    #          make a custom string and maintain the help separately?
    #          the latter is probably easiest...
    #          - Use different background color when displaying the help!
    #
    #        - user can use left/right to see adjacent facts, like normal
    #          (i.e., do not add more widgets/info to carousel!)
    #        - user can adjust adjacent facts' times as well, to
    #          keep pushing facts' times around
    #
    #    X.  - swap 2 adjacent Facts? seems like it makes sense,
    #          but also seems useless, as in, what use case would
    #          have user wanting to swap 2 facts?? i've never wanted
    #          to do that.

    @catch_action_exception
    def edit_time_decrement_start(self, event):
        self.edit_time_adjust(-1, 'start')

    @catch_action_exception
    def edit_time_increment_start(self, event):
        self.edit_time_adjust(1, 'start')

    @catch_action_exception
    def edit_time_decrement_end(self, event):
        self.edit_time_adjust(-1, 'end')

    @catch_action_exception
    def edit_time_increment_end(self, event):
        self.edit_time_adjust(1, 'end')

    @catch_action_exception
    def edit_time_decrement_both(self, event):
        self.edit_time_adjust(-1, 'start', 'end')

    @catch_action_exception
    def edit_time_increment_both(self, event):
        self.edit_time_adjust(1, 'start', 'end')

    def edit_time_adjust(self, delta_mins, *attrs):
        delta_time = self.edit_time_multiplier(delta_mins)
        self.edits_manager.edit_time_adjust(delta_time, *attrs)
        self.zone_manager.reset_diff_fact()
        self.zone_manager.selectively_refresh()

    # Hahaha, (lb): The time_adjust multiplier is a somewhat ridiculous feature.
    #   I'm not sure how useful it is.
    #   Especially if we still need a way for user to edit time directly!

    # (lb): If I hold down a key, I see about a ~ 33 msec. repeat.
    #   Which is probably "hardware settings"-settable per distro.
    #   If I repeatedly manually press a key, the fastest I can do
    #   is about 60 msecs. between presses (though it's more like
    #   100 msec. normally). So we'll try a threshold of 50 msecs.
    #   (Of course, 0.05 doesn't always work. Trying 100 msecs....)
    DOUBLE_KEYLICK_THRESHOLD = 0.10
    MULTIPLY_BY_05_THRESHOLD = 45
    MULTIPLY_BY_10_THRESHOLD = 90
    MULTIPLY_BY_30_THRESHOLD = 150
    MULTIPLY_BY_60_THRESHOLD = 240
    MULTIPLY_BY_1440_THRESHOLD = 375
    MULTIPLY_BY_10080_THRESHOLD = 500
    MULTIPLY_BY_525960_THRESHOLD = 700

    def edit_time_multiplier(self, delta_mins):
        # MEH: (lb): Ideally, we'd hook a global PPT key handler so
        # we could clear the last time_adjust time if a non time_adjust
        # key is pressed; or clear it after some timeout. Whatever; this
        # code has same outcome, even if it doesn't feel ideal.
        if (
            self.last_time_time_adjust is None
            # Be aware of the OS hardware delay before a long press becomes a
            # repeat event, and ignore elapsed time before 1st and 2nd events.
            or (
                self.press_cnt_time_adjust > 1
                and (
                    (time.time() - self.last_time_time_adjust)
                    > UpdateHandler.DOUBLE_KEYLICK_THRESHOLD
                )
            )
        ):
            self.last_time_time_adjust = None
            self.press_cnt_time_adjust = 0
            delta_time = timedelta(minutes=delta_mins * 1)
        elif self.press_cnt_time_adjust < UpdateHandler.MULTIPLY_BY_05_THRESHOLD:
            delta_time = timedelta(minutes=delta_mins * 1)
        elif self.press_cnt_time_adjust < UpdateHandler.MULTIPLY_BY_10_THRESHOLD:
            delta_time = timedelta(minutes=delta_mins * 5)
        elif self.press_cnt_time_adjust < UpdateHandler.MULTIPLY_BY_30_THRESHOLD:
            delta_time = timedelta(minutes=delta_mins * 10)
        elif self.press_cnt_time_adjust < UpdateHandler.MULTIPLY_BY_60_THRESHOLD:
            delta_time = timedelta(minutes=delta_mins * 30)
        elif self.press_cnt_time_adjust < UpdateHandler.MULTIPLY_BY_1440_THRESHOLD:
            delta_time = timedelta(hours=delta_mins)
        elif self.press_cnt_time_adjust < UpdateHandler.MULTIPLY_BY_10080_THRESHOLD:
            delta_time = timedelta(days=delta_mins)
        elif self.press_cnt_time_adjust < UpdateHandler.MULTIPLY_BY_525960_THRESHOLD:
            delta_time = timedelta(weeks=delta_mins)
        else:
            delta_time = timedelta(minutes=delta_mins * 60 * 8760)
        self.last_time_time_adjust = time.time()
        self.press_cnt_time_adjust += 1
        return delta_time

    # ***

    @catch_action_exception
    def edit_fact(self, event):
        """"""
        self.carousel.enduring_edit = True
        event.app.exit()

    def exit_to_awesome_prompt(self, event, restrict_edit):
        self.carousel.enduring_edit = True
        self.carousel.restrict_edit = restrict_edit
        event.app.exit()

    @catch_action_exception
    def edit_actegory(self, event):
        """"""
        self.exit_to_awesome_prompt(event, 'actegory')

    @catch_action_exception
    def edit_tags(self, event):
        """"""
        self.exit_to_awesome_prompt(event, 'tags')

    @catch_action_exception
    def edit_description(self, event):
        """"""
        self.exit_to_awesome_prompt(event, 'description')

    # ***




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Key Binding Action Handler Shim"""

from functools import update_wrapper

from .zone_content import ZoneContent

__all__ = (
    'KeyActionMap',
)


class KeyActionMap(object):
    """"""
    def __init__(self, carousel):
        self.carousel = carousel

        self.zone_manager = carousel.zone_manager

        self.zone_content = carousel.zone_manager.zone_content
        self.zone_details = carousel.zone_manager.zone_details
        self.zone_lowdown = carousel.zone_manager.zone_lowdown

        self.update_handler = carousel.update_handler

    # ***

    class Decorators(object):
        @classmethod
        def debug_log_trace_enter_leave(cls, func):
            def trace_enter_leave_wrapper(obj, event, *args, **kwargs):
                # 2019-01-17: [lb]: I added this wrapper to help delimit debug
                # trace messages (to determine where each command's messages
                # begin and end). But it might later be useful for other tasks,
                # such as profiling. So leaving here, but with a note that says,
                # yeah, this code has little utility to the end consumer, other
                # than to make the developer more comfortable in the code jungle.
                debug = obj.carousel.controller.client_logger.debug
                debug('🚿 🐎 ENTER 👋 🍩 “{}”'.format(func.__name__))
                func(obj, event, *args, **kwargs)
                # Include a visual delimiter to make it easy to scan log trace
                # and see groups of messages belonging to each command.
                debug('🍖 🛀 LEAVE 🐵 🍌 “{}”'.format(func.__name__))

            return update_wrapper(trace_enter_leave_wrapper, func)

    # #### Key bindings wired by key_bonds_normal().

    # ***

    @Decorators.debug_log_trace_enter_leave
    def rotate_help(self, event):
        self.zone_content.rotate_help(event)

    # ***

    @Decorators.debug_log_trace_enter_leave
    def dev_breakpoint(self, event):
        self.carousel.dev_breakpoint(event)

    # ***

    # Next/Prev: Fact

    @Decorators.debug_log_trace_enter_leave
    def jump_fact_dec(self, event):
        self.zone_manager.jump_fact_dec(event)

    @Decorators.debug_log_trace_enter_leave
    def jump_fact_inc(self, event):
        try:
            self.zone_manager.jump_fact_inc(event)
        except Exception:
            # 2019-12-03 01:18: Raised on add-tag then save. Don't remember jumping.
            self.carousel.controller.affirm(False)
            pass

    # Next/Prev: Day

    @Decorators.debug_log_trace_enter_leave
    def jump_day_dec(self, event):
        self.zone_manager.jump_day_dec(event)

    @Decorators.debug_log_trace_enter_leave
    def jump_day_inc(self, event):
        self.zone_manager.jump_day_inc(event)

    # Next/Prev: Rift

    @Decorators.debug_log_trace_enter_leave
    def jump_rift_dec(self, event):
        self.zone_manager.jump_rift_dec(event)

    @Decorators.debug_log_trace_enter_leave
    def jump_rift_inc(self, event):
        self.zone_manager.jump_rift_inc(event)

    # ***

    @Decorators.debug_log_trace_enter_leave
    def cursor_up_one(self, event):
        self.zone_content.cursor_up_one(event)

    @Decorators.debug_log_trace_enter_leave
    def cursor_down_one(self, event):
        self.zone_content.cursor_down_one(event)

    # #### Key bindings wired by key_bonds_save_and_quit().

    @Decorators.debug_log_trace_enter_leave
    def save_edited_and_live(self, event):
        self.carousel.save_edited_and_live(event)

    @Decorators.debug_log_trace_enter_leave
    def save_edited_and_exit(self, event):
        self.carousel.save_edited_and_exit(event)

    @Decorators.debug_log_trace_enter_leave
    def cancel_command(self, event):
        self.carousel.cancel_command(event)

    @Decorators.debug_log_trace_enter_leave
    def cancel_softly(self, event):
        was_helping = self.zone_content.on_reset_hide_help()
        if was_helping:
            return
        self.carousel.cancel_softly(event)

    # ***

    @Decorators.debug_log_trace_enter_leave
    @ZoneContent.Decorators.reset_showing_help
    def ignore_key_press_noop(self, event):
        pass

    # #### Key bindings wired by key_bonds_normal().

    @Decorators.debug_log_trace_enter_leave
    def scroll_up(self, event):
        self.zone_content.scroll_up(event)

    @Decorators.debug_log_trace_enter_leave
    def scroll_down(self, event):
        self.zone_content.scroll_down(event)

    @Decorators.debug_log_trace_enter_leave
    def scroll_top(self, event):
        self.zone_content.scroll_top(event)

    @Decorators.debug_log_trace_enter_leave
    def scroll_bottom(self, event):
        self.zone_content.scroll_bottom(event)

    # ***

    @Decorators.debug_log_trace_enter_leave
    def edit_time_start(self, event):
        self.zone_manager.toggle_focus_time_start(event)

    @Decorators.debug_log_trace_enter_leave
    def edit_time_end(self, event):
        self.zone_manager.toggle_focus_time_end(event)

    @Decorators.debug_log_trace_enter_leave
    def focus_description(self, event):
        self.zone_manager.toggle_focus_description(event)

    # #### Key bindings wired by key_bonds_widget_focus().

    @Decorators.debug_log_trace_enter_leave
    def focus_next(self, event):
        self.zone_manager.focus_next(event)

    @Decorators.debug_log_trace_enter_leave
    def focus_previous(self, event):
        self.zone_manager.focus_previous(event)

    # #### Key bindings wired by key_bonds_update().

    @Decorators.debug_log_trace_enter_leave
    def edit_fact(self, event):
        self.update_handler.edit_fact(event)

    @Decorators.debug_log_trace_enter_leave
    def edit_actegory(self, event):
        self.update_handler.edit_actegory(event)

    @Decorators.debug_log_trace_enter_leave
    def edit_description(self, event):
        self.update_handler.edit_description(event)

    @Decorators.debug_log_trace_enter_leave
    def edit_tags(self, event):
        self.update_handler.edit_tags(event)

    @Decorators.debug_log_trace_enter_leave
    def edit_time_decrement_start(self, event):
        self.update_handler.edit_time_decrement_start(event)

    @Decorators.debug_log_trace_enter_leave
    def edit_time_increment_start(self, event):
        self.update_handler.edit_time_increment_start(event)

    @Decorators.debug_log_trace_enter_leave
    def edit_time_decrement_end(self, event):
        self.update_handler.edit_time_decrement_end(event)

    @Decorators.debug_log_trace_enter_leave
    def edit_time_increment_end(self, event):
        self.update_handler.edit_time_increment_end(event)

    @Decorators.debug_log_trace_enter_leave
    def edit_time_decrement_both(self, event):
        self.update_handler.edit_time_decrement_both(event)

    @Decorators.debug_log_trace_enter_leave
    def edit_time_increment_both(self, event):
        self.update_handler.edit_time_increment_both(event)

    @Decorators.debug_log_trace_enter_leave
    def fact_split(self, event):
        self.update_handler.fact_split(event)

    @Decorators.debug_log_trace_enter_leave
    def fact_erase(self, event):
        self.update_handler.fact_erase(event)

    @Decorators.debug_log_trace_enter_leave
    def fact_merge_prev(self, event):
        self.update_handler.fact_merge_prev(event)

    @Decorators.debug_log_trace_enter_leave
    def fact_merge_next(self, event):
        self.update_handler.fact_merge_next(event)

    # #### Key bindings wired by key_bonds_clipboard().

    @Decorators.debug_log_trace_enter_leave
    def fact_copy_fact(self, event):
        self.update_handler.fact_copy_fact(event)

    @Decorators.debug_log_trace_enter_leave
    def fact_cut(self, event):
        self.update_handler.fact_cut(event)

    @Decorators.debug_log_trace_enter_leave
    def fact_paste(self, event):
        self.update_handler.fact_paste(event)

    @Decorators.debug_log_trace_enter_leave
    def fact_copy_activity(self, event):
        self.update_handler.fact_copy_activity(event)

    @Decorators.debug_log_trace_enter_leave
    def fact_copy_tags(self, event):
        self.update_handler.fact_copy_tags(event)

    @Decorators.debug_log_trace_enter_leave
    def fact_copy_description(self, event):
        self.update_handler.fact_copy_description(event)

    # #### Key bindings wired by key_bonds_undo_redo().

    @Decorators.debug_log_trace_enter_leave
    def undo_command(self, event):
        self.update_handler.undo_command(event)

    @Decorators.debug_log_trace_enter_leave
    def redo_command(self, event):
        self.update_handler.redo_command(event)




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Key Binding Action Manager"""

from .interface_keys import (
    key_bonds_clipboard,
    key_bonds_edit_time,
    key_bonds_normal,
    key_bonds_save_and_quit,
    key_bonds_undo_redo,
    key_bonds_update,
    key_bonds_widget_focus,
    make_bindings
)
from .key_action_map import KeyActionMap

__all__ = (
    'ActionManager',
)


class ActionManager(object):
    """"""
    def __init__(self, carousel):
        self.carousel = carousel

    # ***

    def standup(self):
        self.key_action_map = KeyActionMap(self.carousel)
        self.setup_key_bindings()

    # ***

    def wire_keys_normal(self):
        application = self.carousel.zone_manager.application
        application.key_bindings = self.key_bindings_normal

    def wire_keys_edit_time(self):
        application = self.carousel.zone_manager.application
        application.key_bindings = self.key_bindings_edit_time

    def wire_keys_modal(self):
        application = self.carousel.zone_manager.application
        application.key_bindings = self.key_bindings_modal

    # ***

    def setup_key_bindings(self):
        self.setup_key_bindings_shared()
        self.setup_key_bindings_normal()
        self.setup_key_bindings_edit_time()
        self.setup_key_bindings_modal()

    def setup_key_bindings_shared(self):
        bindings = []
        bindings += key_bonds_save_and_quit(self.key_action_map)
        bindings += key_bonds_widget_focus(self.key_action_map)

        self.key_bindings_shared = bindings

    def setup_key_bindings_normal(self):
        bindings = []
        bindings += key_bonds_normal(self.key_action_map)
        bindings += key_bonds_update(self.key_action_map)
        bindings += key_bonds_clipboard(self.key_action_map)
        bindings += key_bonds_undo_redo(self.key_action_map)
        bindings += self.key_bindings_shared

        self.key_bindings_normal = make_bindings(bindings)

    def setup_key_bindings_edit_time(self):
        bindings = []
        bindings += key_bonds_edit_time(self.carousel.zone_manager.zone_details)
        bindings += key_bonds_undo_redo(self.carousel.zone_manager.zone_details)
        bindings += self.key_bindings_shared

        self.key_bindings_edit_time = make_bindings(bindings)

    def setup_key_bindings_modal(self):
        bindings = []

        self.key_bindings_modal = make_bindings(bindings)




# This file exists within 'dob-viewer':
#
#   https://github.com/hotoffthehamster/dob-viewer
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

"""Key Binding Wiring Manager"""

from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys

from dob_prompt.prompters.interface_bonds import KeyBond

__all__ = (
    'key_bonds_clipboard',
    'key_bonds_edit_time',
    'key_bonds_normal',
    'key_bonds_save_and_quit',
    'key_bonds_undo_redo',
    'key_bonds_update',
    'key_bonds_widget_focus',
    'make_bindings',
)


# ***

def make_bindings(key_bonds):
    key_bindings = KeyBindings()

    for keyb in key_bonds:
        if isinstance(keyb.keycode, tuple):
            key_bindings.add(*keyb.keycode)(keyb.action)
        else:
            key_bindings.add(keyb.keycode)(keyb.action)

    return key_bindings


# ***

def key_bonds_widget_focus(handler):
    key_bonds_widget_focus = [
        KeyBond('tab', action=handler.focus_next),
        KeyBond('s-tab', action=handler.focus_previous),
        # Bindings to edit time are always available (and toggle focus when repeated).
        KeyBond('s', action=handler.edit_time_start),
        KeyBond('e', action=handler.edit_time_end),
    ]
    return key_bonds_widget_focus


# ***

def key_bonds_save_and_quit(handler):
    key_bonds_save_and_quit = [
        # Save Facts command is where you'd expect it.
        KeyBond('c-s', action=handler.save_edited_and_live),
        KeyBond('c-w', action=handler.save_edited_and_exit),
        # User can soft-cancel if they have not edited.
        KeyBond('q', action=handler.cancel_softly),
        # User can always real-quit, but prompted if edits.
        KeyBond('c-q', action=handler.cancel_command),
        # NOTE: Using 'escape' to exit is slow because PPT waits to
        #       see if escape sequence follows (which it wouldn't, after
        #       an 'escape', but meta-combinations start with an escape).
        #   tl;dr: 'escape' to exit slow b/c alias resolution.
        # Note that 'escape' here is the actual ESCape key,
        # and not to be confused with the meta key character,
        # e.g., using ('escape', 'm') to capture Alt-m (m-m).
        KeyBond('escape', action=handler.cancel_softly),
    ]
    return key_bonds_save_and_quit


# ***

def key_bonds_edit_time(handler):
    key_bonds_edit_time = [
        KeyBond('enter', action=handler.editable_text_enter),
        KeyBond('d', action=handler.toggle_focus_description),
        # By default, PPT will add any key we don't capture to active widget's
        # buffer, but we'll override so we can ignore alpha characters.
        KeyBond(Keys.Any, action=handler.editable_text_any_key),
    ]
    return key_bonds_edit_time


# ***

def key_bonds_undo_redo(handler):
    key_bonds_undo_redo = [
        # Vim maps Ctrl-z and Ctrl-y for undo and redo;
        # and u/U to undo count/all and Ctrl-R to redo (count).
        KeyBond('c-z', action=handler.undo_command),
        KeyBond('c-y', action=handler.redo_command),
        # (lb): Too many options!
        # MAYBE: Really mimic all of Vi's undo/redo mappings,
        #        or just pick one each and call it good?
        KeyBond('u', action=handler.undo_command),
        KeyBond('c-r', action=handler.redo_command),
        # (lb): Oh so many duplicate redundant options!
        KeyBond('r', action=handler.redo_command),
    ]
    return key_bonds_undo_redo


# ***

def key_bonds_normal(handler):
    key_bonds_normal = [
        KeyBond('?', action=handler.rotate_help),
        #
        # 2020-03-30: (lb): Was using PPT-HOTH fork and had m- mappings, e.g.,
        #   KeyBond('m-=', action=handler.dev_breakpoint),
        KeyBond((Keys.Escape, '='), action=handler.dev_breakpoint),
        #
        KeyBond('j', action=handler.jump_fact_dec),
        KeyBond('k', action=handler.jump_fact_inc),
        KeyBond(Keys.Left, action=handler.jump_fact_dec),
        KeyBond(Keys.Right, action=handler.jump_fact_inc),
        #
        KeyBond('J', action=handler.jump_day_dec),
        KeyBond('K', action=handler.jump_day_inc),
        # NOTE: It's not, say, 'm-left', but Escape-Arrow.
        # Ahahahaha, alt-arrows are special to Terminator, durp!
        KeyBond((Keys.Escape, Keys.Left), action=handler.jump_day_dec),
        KeyBond((Keys.Escape, Keys.Right), action=handler.jump_day_inc),
        #
        KeyBond(('g', 'g'), action=handler.jump_rift_dec),
        KeyBond('G', action=handler.jump_rift_inc),
        # (lb): Not every Vim key needs to be mapped, e.g.,
        #  KeyBond('M', action=handler.jump_fact_midpoint),
        # seems capricious, i.e., why implement if not just because we can?
        #
        KeyBond('h', action=handler.cursor_up_one),
        KeyBond('l', action=handler.cursor_down_one),
        #
        # FIXME: Oh, maybe make complicated handlers rather than hacking PPT lib?
        # NOTE: If you want to see how key presses map to KeyPresses, try:
        #         $ cd path/to/python-prompt-toolkit/tools
        #         $ python3 debug_vt100_input.py
        #         # PRESS ANY KEY
        # Catch Alt-Up explicitly -- though we don't do anything with it -- so
        # that the 'escape'-looking prefix does not trigger bare 'escape' handler.
        # ((lb): There might be a different way to do this... not sure.)
        KeyBond(  # Alt-UP
            ('escape', '[', '1', ';', '4', 'A'),
            action=handler.ignore_key_press_noop,
        ),
        KeyBond(  # Alt-DOWN
            ('escape', '[', '1', ';', '4', 'B'),
            action=handler.ignore_key_press_noop,
        ),
        KeyBond(  # Alt-RIGHT [bunch of snowflakes]
            ('escape', '[', '1', ';', '4', 'C'),
            action=handler.ignore_key_press_noop,
        ),
        KeyBond(  # Alt-LEFT
            ('escape', '[', '1', ';', '4', 'D'),
            action=handler.ignore_key_press_noop,
        ),
        #
        KeyBond('pageup', action=handler.scroll_up),
        KeyBond('pagedown', action=handler.scroll_down),
        KeyBond('home', action=handler.scroll_top),
        KeyBond('end', action=handler.scroll_bottom),
        #
        # FIXME/BACKLOG: Search feature. E.g., like Vim's /:
        #   KeyBond('/', action=zone_lowdown.start_search),
        # FIXME/BACKLOG: Filter feature.
        #   (By tag; matching text; dates; days of the week; etc.)
    ]
    return key_bonds_normal


# ***

def key_bonds_update(handler):
    key_bonds_update = [
        # Raw Fact Editing. With a Capital E.
        KeyBond('E', action=handler.edit_fact),
        # Edit act@gory, description, and tags using prompt__awesome.
        # (lb): This is pretty cool. prompt_awesome was built first,
        # and then I got comfortable with PPT and built the Carousel,
        # and then I was able to stick one inside the other, and it's
        # just awesome awesome now.
        KeyBond('a', action=handler.edit_actegory),
        KeyBond('d', action=handler.edit_description),
        KeyBond('t', action=handler.edit_tags),
        #
        KeyBond('s-left', action=handler.edit_time_decrement_start),
        KeyBond('s-right', action=handler.edit_time_increment_start),
        KeyBond('c-left', action=handler.edit_time_decrement_end),
        KeyBond('c-right', action=handler.edit_time_increment_end),
        # FIXME/2019-01-21: Can you check if running in Terminator
        #  and warn-tell user. And/Or: Customize Key Binding feature.
        # In Terminator: Shift+Ctrl+Left/+Right: Resize the terminal left/right.
        #  (lb): I've disabled the 2 bindings in Terminator,
        #   so this works for me... so fixing it is a low priority!
        KeyBond('s-c-left', action=handler.edit_time_decrement_both),
        KeyBond('s-c-right', action=handler.edit_time_increment_both),
        #
        # 2020-03-30: (lb): Was using PPT-HOTH fork and had m- mappings, e.g.,
        #   KeyBond('m-p', action=handler.fact_split),
        #   KeyBond('m-e', action=handler.fact_erase),
        #   KeyBond(('m-m', Keys.Left), action=handler.fact_merge_prev),
        #   KeyBond(('m-m', Keys.Right), action=handler.fact_merge_next),
        KeyBond((Keys.Escape, 'p'), action=handler.fact_split),
        KeyBond((Keys.Escape, 'e'), action=handler.fact_erase),
        KeyBond((Keys.Escape, 'm', Keys.Left), action=handler.fact_merge_prev),
        KeyBond((Keys.Escape, 'm', Keys.Right), action=handler.fact_merge_next),
    ]
    return key_bonds_update


def key_bonds_clipboard(handler):
    key_bonds_clipboard = [
        #
        KeyBond('c-c', action=handler.fact_copy_fact),
        KeyBond('c-x', action=handler.fact_cut),
        KeyBond('c-v', action=handler.fact_paste),
        #
        KeyBond(('A', 'c-c'), action=handler.fact_copy_activity),
        KeyBond(('T', 'c-c'), action=handler.fact_copy_tags),
        KeyBond(('D', 'c-c'), action=handler.fact_copy_description),
    ]
    return key_bonds_clipboard

